/* Mergesort su liste. */
#include <stdio.h>
#include <stdlib.h>

typedef struct ele {
  int chiave;
  struct ele *prox;
} elemento;

/* Inserisce in testa. */
elemento *insTesta( elemento *testa, int v )
{
  elemento *nuovo = malloc(sizeof(elemento));
  nuovo->chiave = v;
  nuovo->prox = testa;
  return nuovo;
}


/* Fusione di due liste. */
elemento *fondi ( elemento *t1, elemento *t2 )
{
   /* Manteniamo una testa della nuova lista,
      una coda della nuova lista, una variabile temporanea.*/
   elemento *nuovalista = NULL;
   elemento *codalista = NULL;
   elemento *temp;

   /* Il codice e' simile alla classica fusione, ma 
      non richiede un array di supporto. */
   /* Il primo ciclo e' eseguito finche' entrambe
      le liste sono non-vuote. */
   while ( t1 != NULL && t2 != NULL )
   {
      /* Scegliamo quale chiave da copiare,
         e salviamo l'elemento, subito dopo 
         andando sul successore. Poi mettiamo
         l'elemento salvato in coda alla nuova
         lista. */
      if ( t1->chiave < t2->chiave ) 
      {
         temp = t1;
         t1 = t1->prox;
      } else { 
         temp = t2;
         t2 = t2->prox;
      }
      /* Facciamo attenzione alla lista vuota. */
      if ( nuovalista == NULL ) 
            codalista = nuovalista = temp;
      else
      {
          codalista->prox = temp;
          codalista = codalista->prox;
       } 
   }

   /* Se la prima lista non e' terminata,
      la concateniamo. */
   if ( t1 != NULL ) 
   {
     if (nuovalista == NULL ) 
        nuovalista = t1;
     else
        codalista->prox = t1;
   }
   /* Se la seconda lista non e' terminata,
      invece, concateniamo lei. */
   else if ( t2 != NULL ) {
     if ( nuovalista == NULL ) 
        nuovalista = t2;
     else
        codalista->prox = t2;
   }
   
   /* Ritorniamo la testa della nuova lista. */
   return nuovalista; 
}

elemento * trova_meta_prev2 ( elemento *testa ) 
{
    elemento *p = testa;
    int len = 0;

    while ( p != NULL ) 
    {
      p = p->prox;
      ++len;
    }
    /* Len e' sempre almeno 2. */
    len  = len / 2 - 1;
    p = testa;
    while ( len -- != 0 ) 
      p = p->prox;
    return p;
}

/* Trova l'ultimo elemento prima di dove le due liste vanno
   divise, facendo l'assunzione che la lista in input abbia
   almeno 2 elementi detntro.
   Il modo clasico (conta la dimensione della lista
   e poi ri-scandiscila per meta' -1 ) e' descritto nella
   funzione "trova_meta_prev2" sopra.
   Questo metodo, a mio parere piu' corto, elegante, e meno
   prono ad errori, e' diverso. Si parte con 2 puntatori all'
   inizio della lista. Uno salta di 2 elementi alla volta, 
   mentre l'altro di 1 elemento alla volta, fino a quando il primo
   non si trova sulla coda. In questo modo, il primo si trovera'
   a meta' esatta della lista, meno 1.
*/
elemento *trova_meta_prev(elemento *lista) 
{
   elemento *doppio = lista, *singolo = lista;
   while ( doppio != NULL && doppio->prox != NULL ) 
   {
      doppio = doppio->prox;
      if ( doppio->prox ==  NULL ) break;
      doppio = doppio->prox;
      singolo = singolo->prox;
   }
   return singolo;
}

void stampa(elemento *l )
{
   while ( l != NULL )
   {
     printf("%d ", l->chiave);
     l = l->prox;
   }
   printf("\n");
}

/* Mergesort su liste:
   trova la meta', spacca in due la lista,
   ordina la parte sinistra, ordina la parte destra,
   fondi. */
elemento *msort(elemento *testa)
{
    int i;
    elemento *p, *mezzo;
 
    /* Lista vuota o singola */
    if ( testa == NULL || testa->prox == NULL ) 
         return testa;

     /* Troviamo l'elemento precedente alla meta' */
     p = trova_meta_prev ( testa );
     /* Spezziamo la lista in due, tenendoci la  
        meta' come seconda testa. */
     mezzo = p->prox; 
     p->prox = NULL; 
     /* Ordiniamo le due meta' */
       
     testa = msort ( testa );
     mezzo = msort ( mezzo );
     /* A questo punto fondiamo */
     return fondi ( testa, mezzo );
}



main()
{
   int N, i, val;
   elemento *lista = NULL;

   scanf("%d", &N);
   for(i = 0; i < N; i++ )
   {
     scanf("%d", &val );
     /* Tanto gli oggetti vanno ordinati,
        per semplificare le cose, inseriamo
        in testa. */
     lista = insTesta ( lista,val );
   } 

   stampa(lista);

   lista = msort(lista);
   
   stampa(lista);
   return 0;
}
